-- gategen.vhdl
-- created : jeu. août  8 10:18:01 CEST 2019  by Yann Guidon (whygee@f-cpu.org)
-- version : sam. août 10 08:06:58 CEST 2019
-- version : dim. août 11 17:36:51 CEST 2019
--
-- Released under the GNU AGPLv3 license
--
-- This file generates PA3_components.vhdl and PA3_definitions.vhdl.


Library ieee;
    use ieee.std_logic_1164.all;
    use ieee.numeric_std.all;
Library work;
    use work.all;
    use work.PA3_genlib.all;
use STD.textio.all;

entity GateGen is
  generic ( gate_delay : string := "";
            fast : string := "";
            dummy_str : string := "");
end GateGen;

architecture runme of GateGen is
  file gate_defs : text open write_mode is "PA3_definitions.vhdl";
  file gate_list : text open write_mode is "PA3_components.vhdl";
begin

  bench: process
    variable c, Epol, Cpol, force : character;
    variable i, j : integer := 0;
    variable s : EnumGates2;
    variable t : EnumGates3;
    variable L8 : LookupType8;
    variable L4 : LookupType4;
    variable op : character;
    variable row : line;
    variable Str2 : string(1 to 2);
    variable Str5 : string(1 to 5);

    procedure wl(s: string) is
    begin
      write(row, s);
      writeline(gate_list, row);
    end wl;

    procedure wd(s: string) is
    begin
      write(row, s);
      writeline(gate_defs, row);
    end wd;

    procedure uselib is
    begin
      wd("");
      wd("--------------------------------------------------");
      wd("");
      wd("Library ieee;");
      wd("    use ieee.std_logic_1164.all;");
      wd("Library work;");
      wd("    use work.all;");
      wd("    use work.PA3_genlib.all;");
      wd("");
    end uselib;

    procedure census is
    begin
      if fast = "" then
        wd("  generic( Gate_Number : integer := Gate_Census(0));");
      end if;
    end census;

  begin
    wl("-- File : PA3_components.vhdl");
    wl("-- DO NOT MODIFY ! Generated file !");
    wl("-- (C) Yann Guidon 20190808 - 20190810");
    wl("");
    wl("Library ieee;");
    wl("    use ieee.std_logic_1164.all;");
    wl("");
    wl("package proasic3 is");

    wd("-- File : PA3_definitions.vhdl");
    wd("-- DO NOT MODIFY ! Generated file !");
    wd("-- (C) Yann Guidon 20190808 - 20190810");
    if fast /= "" then
      wd("-- FAST/lean version");
      report " gategen : generating FAST/lean version";
    else
      wd("-- Instrumented version");
      report " gategen : generating instrumented version";
    end if;
    if gate_delay = "" then
      report "No gate delay specified. Change this value with -ggate_delay=""1 ns"" ";
    else
      report "Gate delay set to " & gate_delay;
    end if;

    wl("");
    wl("  component register_generics port(gate_select_number, ");
    wl("bit_select_number: in integer; clock : in std_logic); end component;"); 
    wl("");

    uselib;
    wd("entity register_generics is");
    wd("  port( gate_select_number: in integer;");
    wd("         bit_select_number: in integer;");
    wd("        clock : in std_logic);");
    wd("end register_generics;");
    wd("");
    wd("architecture rtl of register_generics is");
    wd("begin");
    if fast = "" then
      wd("  update_select_gate(gate_select_number,"
          & " bit_select_number); -- called only once at start");
      wd("");
      wd("  process(clock) begin");
      wd("    case clock is");
      wd("      when '1' => update_histogram;");
      wd("      when 'X' => display_histogram;");
      wd("      when others => NULL;");
      wd("    end case;");
      wd("  end process;");
    end if;
    wd("end rtl;");

    wl("-- 0-input gates :");

    for i in NoInputGates'range loop
      wl("  component "& NoInputGates(i) & " port(Y : out std_logic); end component;");

      uselib;
      wd("entity " & NoInputGates(i) & " is");
      census;
      wd("  port(Y : out std_logic);");
      wd("end " & NoInputGates(i) & ";");
      wd("");
      wd("architecture rtl of " & NoInputGates(i) & " is");
      if fast = "" then
        wd("  signal LUT1 : LookupType1;");
        wd("begin");
        wd("  LUT1 <= InferLUT(""" & integer'image(i) & """, LUT1'instance_name, Gate_Number, 0);");
        wd("  Y <= LUT1(0);");
      else
        wd("begin");
        wd("  Y <= '" & integer'image(i) & "';");
      end if;
      wd("end rtl;");
    end loop;

    wl("");
    wl("-- 1-input gates :");


    wl("");
    wl("-- 2-inputs gates :");

    for s in EnumGates2 loop
      L4 := ComputeLUT4(s);
      write(row, "  component " & EnumGates2'image(s), left, 19);
      wl(" port(A, B : in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L4));

      uselib;
      wd("entity " & EnumGates2'image(s) & " is");
      census;
      wd("  port( A, B : in std_logic;");
      wd("           Y : out std_logic);");
      wd("end " & EnumGates2'image(s) & ";");
      wd("");
      wd("architecture rtl of " & EnumGates2'image(s) & " is");
      if fast = "" then
        wd("  signal LUT4 : LookupType4;");
        wd("begin");
        wd("  LUT4 <= InferLUT(""" & SLV_to_bin(L4)
           & """, LUT4'instance_name, Gate_Number, 3);");
        write(row, "  Y <= LookItUp2(A, B, LUT4, Gate_Number)" & fast);  -- kludge, fast=""
      else
        wd("begin");
        write(row, "  Y <= LookItUp2_fast(A, B, """ & SLV_to_bin(L4) & """)");
      end if;

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");

      wd("end rtl;");
    end loop;

    wl("");
    wl("-- 3-inputs gates :");
    for t in EnumGates3 loop
      case t is
        when MX2  |
             MX2A |
             MX2B |
             MX2C   =>  op := 'S';
        when others =>  op := 'C';
      end case;

      L8 := ComputeLUT8(t);
      write(row, "  component " & EnumGates3'image(t), left, 19);
      wl(" port(A, B, " & op
          & " : in std_logic; Y : out std_logic); end component; -- " & SLV_to_bin(L8));

      uselib;
      wd("entity " & EnumGates3'image(t) & " is");
      census;
      wd("  port( A, B, " & op &" : in std_logic;");
      wd("              Y : out std_logic);");
      wd("end " & EnumGates3'image(t) & ";");
      wd("");
      wd("architecture rtl of " & EnumGates3'image(t) & " is");

      if fast = "" then
        wd("  signal LUT8 : LookupType8;");
        wd("begin");
        wd("  LUT8 <= InferLUT(""" & SLV_to_bin(L8)
            & """, LUT8'instance_name, Gate_Number, 7);");    -- called only once at start
        write(row, "  Y <= LookItUp3(A, B, " & op & ", LUT8, Gate_Number)" & dummy_str);  -- kludge
      else
	wd("begin");
        write(row, "  Y <= LookItUp3_fast(A, B, " & op &", """ & SLV_to_bin(L8) & """)");
      end if;

      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");
      wd("end rtl;");

    end loop;

    wl("");
    wl("-- Sequential gates :");

    for i in EnumDFF'range loop
      str2 := "  ";
      j := 5;
      if EnumDFF(i)(5) = 'E' then
        str2 := "E,";
        j := 7;
        Epol := EnumDFF(i)(6);
      end if;

      str5 := "     ";
      if EnumDFF(i)(j) = 'P' then
        str5 := ", PRE";
        force := '1';
        Cpol := EnumDFF(i)(j+1);
      else
        if EnumDFF(i)(j) = 'C' then
          str5 := ", CLR";
          force := '0';
          Cpol := EnumDFF(i)(j+1);
        end if;
      end if;

      wl("  component " & EnumDFF(i) & " port(D, " & str2 & " CLK" & str5
         & " : in std_logic; Q : out std_logic); end component;");

      uselib;
      wd("entity " & EnumDFF(i) & " is");
      wd("  port(D, " & str2 & " CLK" & str5 & " : in std_logic;");
      wd("       Q : out std_logic);");
      wd("end " & EnumDFF(i) & ";");
      wd("");
      wd("architecture rtl of " & EnumDFF(i) & " is");
      wd("begin");
      write(row, "  process(CLK" & dummy_str);
      if str5(1) = ',' then
        write(row, str5);
      end if;
      wd(")");
      wd("  begin");

      if str5(1) = ',' then
        wd("    if " & str5(3 to 5) & "='" & Cpol & "' then");
        write(row, "      Q <= '" & force & "'");
        if gate_delay /= "" then
          write(row, " after " & gate_delay);
        end if;
        wd(";");
        wd("    else");
      end if;

      write(row, "      if rising_edge(CLK)" & dummy_str);
      if str2(1) = 'E' then
        write(row, " and E='" & Epol & "'");
      end if;
      wd(" then");
      write(row, "        Q <= D " & dummy_str);
      if gate_delay /= "" then
        write(row, " after " & gate_delay);
      end if;
      wd(";");
      wd("      end if;");

      if str5(1) = ',' then
        wd("    end if;");
      end if;

      wd("  end process;");
      wd("end rtl;");

    end loop;

    wl("");
    wl("end proasic3;");

    wait;
  end process;

end runme;
