-- file test_gates.vhdl
-- created ven. août  9 03:14:28 CEST 2019
-- version sam. août 10 16:08:18 CEST 2019
-- version dim. août 11 17:36:51 CEST 2019
--
-- Released under the GNU AGPLv3 license
--
-- This is a simple/small example that shows how to use
-- the extra features of the rewritten proasic3 library.

Library ieee;
    use ieee.std_logic_1164.all;
Library proasic3;
    use proasic3.all;

entity test_gates is
  generic (
    gate_select_number: integer := 0;
    bit_select_number: integer := -1;
    disturb: integer := 0
  );
end test_gates;

architecture rtl of test_gates is
  signal A, B, C, D, E, F, G, Y : std_logic;
  signal clock : std_logic := '0';
begin
  -- this entity MUST be written first,
  -- at the very beginning of the toplevel file:
  rg: entity register_generics port map(
       gate_select_number => gate_select_number,
       bit_select_number => bit_select_number,
       clock => clock);

  x1: entity xor2  port map(A => A, B=>B,       Y=>E );
  x2: entity xor2  port map(A => C, B=>E,       Y=>F );
  x3: entity GND   port map(                    Y=>G );
  x4: entity xor3  port map(A => D, B=>F, C=>G, Y=>Y );

  test: process
    variable i : integer := 0;
  begin
    if gate_select_number=-1 then
      wait;
    end if;

    wait for 100 ns;

    A <= '0';
    loop

      B <= '0';
      loop

	C <= '0';
        loop

          D <= '0';
          loop

            i := i+1;
            if i = 9 and disturb /= 0 then
              B <= 'L';
            end if;

            clock <= '0';
            wait for 10 ns;
            clock <= '1';
            wait for 10 ns;
            report  " A=" & std_logic'image(A) &
                    " B=" & std_logic'image(B) &
                    " C=" & std_logic'image(C) &
                    " D=" & std_logic'image(D) &
                    " E=" & std_logic'image(E) &
                    " F=" & std_logic'image(F) &
                    " G=" & std_logic'image(G) &
                    " Y=" & std_logic'image(Y);

            exit when D = '1';
            D <= '1';
          end loop;

          exit when C = '1';
          C <= '1';
        end loop;

        exit when B = '1';
        B <= '1';
      end loop;

      exit when A = '1';
      A <= '1';
    end loop;

    clock<='X'; -- display the histogram
    wait;
  end process;

end rtl;
